"""
##################################################################################
#
# Website As App
# Run any website as standalone desktop application
#
# @author    Marcin Orlowski <mail (#) marcinOrlowski (.) com>
# @copyright 2023-2024 Marcin Orlowski
# @license   https://www.opensource.org/licenses/mit-license.php MIT
# @link      https://github.com/MarcinOrlowski/website-as-app
#
# @file      websiteapp/webapp.py
#
##################################################################################
"""
import os
import sys
from typing import Optional

import fasteners
from PyQt5.QtWebEngineWidgets import QWebEngineView
from PySide6.QtCore import QUrl, QFileSystemWatcher, Qt
from PySide6.QtGui import QAction
from PySide6.QtWebEngineCore import QWebEnginePage
from PySide6.QtWebEngineCore import (
    QWebEngineProfile,
    QWebEngineSettings,
)
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QVBoxLayout,
    QWidget,
    QSystemTrayIcon,
    QMenu,
    QFileDialog,
)

from websiteapp.about import About
from websiteapp.const import Const
from websiteapp.utils import Utils
from websiteapp.webengine import CustomWebEngineView


class WebApp(QMainWindow):
    about_dialog: Optional[About] = None
    lock: Optional[fasteners.InterProcessLock] = None
    file_watcher: Optional[QFileSystemWatcher] = None

    def __init__(self):
        super().__init__()

        # Check for --version before full argument parsing
        if '--version' in sys.argv:
            print(f'{Const.APP_NAME} {Const.APP_VERSION}')
            sys.exit(0)

        self.app = QApplication.instance()
        self.args = Utils.handle_args()

        if not self.args.allow_multiple:
            if not self.acquire_lock():
                sys.exit(0)  # Exit silently as we've activated the existing instance

        self.setup_activation_listener()

        # Set window geometry
        x, y, width, height = Utils.parse_geometry(self.args.geometry)
        self.setGeometry(x, y, width, height)
        self.dbug(f'Geometry: {width}x{height}+{x}+{y}')

        app_icon = Utils.get_icon(self.args.icon)
        self.setWindowIcon(app_icon)
        if not self.args.no_tray:
            self.setup_tray_icon(app_icon)

        window_title = self.args.name if self.args.name else self.args.url
        window_title += f' ({self.args.profile})' if self.args.debug else ''
        self.setWindowTitle(f'{window_title} · {Const.APP_NAME}')

        # Create a persistent profile (cookie jar etc.)
        self.dbug(f'Profile: {self.args.profile}')
        self.profile = QWebEngineProfile(self.args.profile, self)
        self.page = QWebEnginePage(self.profile, self)

        # Create and configure the webpage
        self.page = QWebEnginePage(self.profile, self)
        # Connect permission request handler
        self.page.featurePermissionRequested.connect(self.handle_permission_request)

        # Handle downloads
        self.profile.downloadRequested.connect(self.on_download_requested)

        # self.browser = QWebEngineView(self)
        # Modified to use CustomWebEngineView
        if self.args.no_custom_webengine:
            self.browser = QWebEngineView(self)
        else:
            self.browser = CustomWebEngineView(self, debug=self.args.debug)
        self.browser.setPage(self.page)
        self.browser.setZoomFactor(self.args.zoom)

        web_settings = self.browser.settings()

        # Enable all required clipboard permissions
        web_settings.setAttribute(QWebEngineSettings.WebAttribute.JavascriptCanAccessClipboard,
                                  True)
        web_settings.setAttribute(QWebEngineSettings.WebAttribute.JavascriptCanPaste, True)
        web_settings.setAttribute(
            QWebEngineSettings.WebAttribute.AllowWindowActivationFromJavaScript, True)
        # Additional profile settings for clipboard
        self.profile.settings().setAttribute(
            QWebEngineSettings.WebAttribute.JavascriptCanAccessClipboard, True)
        self.profile.settings().setAttribute(QWebEngineSettings.WebAttribute.JavascriptCanPaste,
                                             True)

        self.browser.setPage(self.page)
        self.browser.setZoomFactor(self.args.zoom)

        # Ensure the browser widget can receive focus and key events
        self.browser.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.browser.setFocus()

        self.dbug(f'URL: {self.args.url}')
        self.browser.setUrl(QUrl(self.args.url))

        # Window layout
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.browser)

        central_widget = QWidget()
        central_widget.setLayout(layout)
        self.setCentralWidget(central_widget)

    def acquire_lock(self) -> bool:
        """
        Acquires a lock for the current profile to prevent multiple instances.
        Returns True if lock was acquired, False otherwise.
        """
        lock_file = os.path.join(os.path.expanduser("~"), f".websiteapp_{self.args.profile}.lock")
        self.lock = fasteners.InterProcessLock(lock_file)
        if not self.lock.acquire(blocking=False):
            self.activate_existing_instance()
            return False
        return True

    def activate_existing_instance(self) -> None:
        """
        Sends a signal to the existing instance to activate its window.
        """
        self.dbug(f"Bringing existing instance: {Const.APP_NAME}, profile: {self.args.profile}")
        signal_file = os.path.join(os.path.expanduser("~"),
                                   f".websiteapp_{self.args.profile}_signal")
        with open(signal_file, 'w') as f:
            f.write("activate")

    def setup_activation_listener(self) -> None:
        """
        Sets up a file system watcher to listen for activation signals.
        """
        self.file_watcher = QFileSystemWatcher(self)
        signal_file = os.path.join(os.path.expanduser("~"),
                                   f".websiteapp_{self.args.profile}_signal")
        self.file_watcher.addPath(os.path.dirname(signal_file))
        self.file_watcher.directoryChanged.connect(self.check_activation_signal)

    def check_activation_signal(self, path):
        """
        Checks for activation signals and brings the window to front if signal is received.
        """
        signal_file = os.path.join(path, f".websiteapp_{self.args.profile}_signal")
        if os.path.exists(signal_file):
            os.remove(signal_file)
            self.activate_window()

    def activate_window(self) -> None:
        """
        Brings the window to front and restores it if minimized.
        """
        self.showNormal()
        self.activateWindow()
        self.raise_()

    def closeEvent(self, event) -> None:
        """
        Override closeEvent to hide the window instead of closing it and hides the window instead.

        :param event:
        """
        if not self.args.no_tray:
            self.hide()
            event.ignore()
        else:
            if self.lock:
                self.lock.release()
                self.lock = None
            super().closeEvent(event)

    def setup_tray_icon(self, icon) -> None:
        """
        This method is used to set up the system tray icon for the application.

        :param icon: QPixmap object representing the icon that will be used in the system tray.
        """
        tray_menu = QMenu()

        tray_icon = QSystemTrayIcon(icon, self.app)
        tray_icon.setContextMenu(tray_menu)
        tray_icon.activated.connect(self.on_tray_icon_activated)

        about_label = f'About {Const.APP_NAME}'
        about_action = QAction(about_label, self.app)
        about_action.triggered.connect(self.open_about_dialog)
        tray_menu.addAction(about_action)

        show_label = f'Show {self.args.name}' if self.args.name else 'Show'
        show_action = QAction(show_label, self)
        show_action.triggered.connect(self.show)
        tray_menu.addAction(show_action)

        quit_label = f'Quit {self.args.name}' if self.args.name else 'Quit'
        quit_action = QAction(quit_label, self.app)
        quit_action.triggered.connect(self.quit_app)
        tray_menu.addAction(quit_action)

        tray_icon.show()

    def quit_app(self) -> None:
        """
        Closes the application.
        """
        if self.lock:
            self.lock.release()
            self.lock = None
        self.app.quit()

    def open_about_dialog(self) -> None:
        """
        Opens the about dialog.
        """
        if not self.about_dialog:
            self.about_dialog = About()

        if not self.about_dialog.isVisible():
            self.about_dialog.show()
        self.about_dialog.activateWindow()

    def on_tray_icon_activated(self, reason) -> None:
        """
        Callback invoked when application icon in system tray is clicked.
        """
        if reason == QSystemTrayIcon.ActivationReason.Trigger:
            # Check for left-click (Trigger)
            self.toggle_window()

    def toggle_window(self) -> None:
        """
        Toggles the visibility of the window.
        """
        self.hide() if self.isVisible() else self.show()

    # ############################################################################################ #

    def dbug(self, msg: str) -> None:
        if self.args.debug:
            print(msg, file=sys.stderr)

    # ############################################################################################ #

    @staticmethod
    def run() -> None:
        """
        Application entry point. When renamed, ensure setup.py's reference is updated as well.
        """
        app = QApplication(sys.argv)
        app.setOrganizationName("MarcinOrlowski")
        app.setApplicationName("Website As App")

        window = WebApp()

        if window.args.minimized and window.args.no_tray:
            # Cannot use --no-tray and --minimized at the same time. Ignoring --minimized
            window.args.minimized = False

        if not window.args.minimized:
            window.show()

        exit_code = app.exec()

        if window.lock:
            window.lock.release()
            window.lock = None

        sys.exit(exit_code)

    def on_download_requested(self, download) -> None:
        """
        Handles file download requests from the web page.
        """
        print('on_download_requested')
        # Prompt the user to select a download location
        suggested_filename = download.downloadFileName()
        options = QFileDialog.Options()
        path, _ = QFileDialog.getSaveFileName(self, "Save File", suggested_filename,
                                              options=options)
        if path:
            download.setDownloadFileName(os.path.basename(path))
            download.setDownloadDirectory(os.path.dirname(path))
            download.accept()
        else:
            download.cancel()

    def handle_permission_request(self, origin, feature) -> None:
        """
        Handle permission requests from the webpage.
        """
        from PySide6.QtWebEngineCore import QWebEnginePage

        self.dbug(f"Permission requested: {feature} from {origin}")

        # Define all clipboard-related features
        clipboard_features = [
            QWebEnginePage.Feature.ClipboardReadWrite,
            QWebEnginePage.Feature.Clipboard,  # For backwards compatibility
        ]

        if feature in clipboard_features:
            self.dbug(f"Granting clipboard permission for feature: {feature}")
            self.page.setFeaturePermission(
                origin,
                feature,
                QWebEnginePage.PermissionPolicy.PermissionGrantedByUser
            )
        else:
            self.dbug(f"Denying permission for feature: {feature}")
            self.page.setFeaturePermission(
                origin,
                feature,
                QWebEnginePage.PermissionPolicy.PermissionDeniedByUser
            )



"""
##################################################################################
#
# Website As App
# Run any website as standalone desktop application
#
# @author    Marcin Orlowski <mail (#) marcinOrlowski (.) com>
# @copyright 2024 Marcin Orlowski
# @license   https://www.opensource.org/licenses/mit-license.php MIT
# @link      https://github.com/MarcinOrlowski/website-as-app
#
# @file      websiteapp/webengine.py
#
##################################################################################
"""

# Target file: websiteapp/custom_web_view.py

import re
from PySide6.QtWidgets import QMenu, QApplication, QMessageBox
from PySide6.QtGui import QAction
from PySide6.QtWebEngineCore import QWebEnginePage
from PySide6.QtWebEngineWidgets import QWebEngineView

class CustomWebEngineView(QWebEngineView):
    def __init__(self, parent=None, debug=False):
        super().__init__(parent)
        self.debug = debug  # Store the debug flag to conditionally add the dump action
        self.loading = False  # Keep track of whether a page is loading

        # Connect load progress to determine loading status
        self.loadStarted.connect(self.on_load_started)
        self.loadFinished.connect(self.on_load_finished)

    def on_load_started(self):
        """
        Slot called when page load starts.
        """
        self.loading = True

    def on_load_finished(self):
        """
        Slot called when page load finishes.
        """
        self.loading = False

    def contextMenuEvent(self, event):
        menu = QMenu(self)

        # Add default actions to the context menu using WebAction namespace
        actions = [
            QWebEnginePage.WebAction.Back,
            QWebEnginePage.WebAction.Forward,
            QWebEnginePage.WebAction.Reload,
            QWebEnginePage.WebAction.Stop,
            QWebEnginePage.WebAction.Copy,
            QWebEnginePage.WebAction.Cut,
            QWebEnginePage.WebAction.Paste,
            QWebEnginePage.WebAction.SelectAll
        ]

        for action in actions:
            q_action = self.pageAction(action)
            if action == QWebEnginePage.WebAction.Back:
                # Specifically handle "Back" item to ensure it's enabled/disabled correctly
                q_action.setText("Back")
                # Check if the back action should be disabled
                history = self.history()
                if len(history.backItems(history.count())) <= 1:
                    q_action.setEnabled(False)

            if action == QWebEnginePage.WebAction.Stop:
                # Disable "Stop" action if there is nothing to stop
                if not self.loading:
                    q_action.setEnabled(False)

            # if q_action.isEnabled():  # Only add enabled actions
            menu.addAction(q_action)

        # Add custom action to copy URL
        copy_url_action = QAction("Copy Current URL", self)
        copy_url_action.triggered.connect(self.copy_url_to_clipboard)
        menu.addAction(copy_url_action)

        # Add custom action to open URL from clipboard
        paste_url_action = QAction("Open URL from Clipboard", self)
        paste_url_action.triggered.connect(self.paste_url_from_clipboard)

        # Check if clipboard contains a valid URL
        if self.is_valid_url_in_clipboard():
            paste_url_action.setEnabled(True)
        else:
            paste_url_action.setEnabled(False)

        menu.addAction(paste_url_action)

        # # If debug mode is enabled, add the "Dump Back Stack" option
        # if self.debug:
        #     dump_stack_action = QAction("Dump Back Stack", self)
        #     dump_stack_action.triggered.connect(self.dump_back_stack)
        #     menu.addAction(dump_stack_action)

        # Add custom action to quit the application
        quit_action = QAction("Quit app…", self)
        quit_action.triggered.connect(self.quit_app)
        menu.addAction(quit_action)

        # Display the unified context menu at the cursor position
        menu.exec(event.globalPos())

    def copy_url_to_clipboard(self):
        """
        Copies the current page URL to the clipboard.
        """
        current_url = self.url().toString()
        clipboard = QApplication.clipboard()
        clipboard.setText(current_url)

    def paste_url_from_clipboard(self):
        """
        Pastes a URL from the clipboard into the web client if valid. Otherwise, shows an error dialog.
        """
        clipboard = QApplication.clipboard()
        clipboard_text = clipboard.text()

        if self.is_valid_url(clipboard_text):
            self.setUrl(clipboard_text)
        else:
            QMessageBox.warning(self, "Invalid URL", "No valid URL in clipboard found.")

    def dump_back_stack(self):
        """
        Dumps the list of all pages that can be navigated back to.
        """
        history = self.history()
        back_items = history.backItems(history.count())
        print("Back Stack:")
        for item in back_items:
            print(f"  - {item.url().toString()}")

    def quit_app(self):
        """
        Quits the application gracefully.
        """
        QApplication.quit()

    def is_valid_url_in_clipboard(self):
        """
        Checks if the clipboard contains a valid URL.
        """
        clipboard = QApplication.clipboard()
        clipboard_text = clipboard.text()
        return self.is_valid_url(clipboard_text)

    @staticmethod
    def is_valid_url(url):
        """
        Validates if the given string is a valid URL.
        """
        url_pattern = re.compile(
            r'^(https?|ftp)://'  # http:// or https:// or ftp://
            r'(\w+(\-\w+)*\.)+[a-zA-Z]{2,}'  # domain name
            r'(:\d+)?'  # optional port
            r'(\/[^\s]*)?$'  # optional path
        )
        return bool(url_pattern.match(url))
